/*
    pwl9x "Windows 9x PassWord List reader"
    Copyright (C) 2001 Knispel Guillaume

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"
#include "cpu.h"
#include "mmx.h"
#include "x86.h"

void (*try_crack_pwl)(Params *p, Pwl *pwl);

void chose_crack_core(Params *p)
{
  Cpu_desc cpu;
  char *vendorstr, *feat;
  
  idcpu(&cpu);
  switch(cpu.vendid) {
  case INTEL: vendorstr = "n Intel"; break;
  case AMD: vendorstr = "n Amd"; break;
  case IDT: vendorstr = "n Idt"; break;
  case NG: vendorstr = " NexGen"; break;
  case CYRIX: vendorstr = " Cyrix"; break;
  case UMC: vendorstr = "n Umc"; break;
  case RISE: vendorstr = " Rise"; break;
  case TRANSM: vendorstr = " Transmeta"; break;
  default: vendorstr = "n unknown"; break;
  }

  printf("Detected a%s processor, family %d\n  l1 code cache size = %d Kb\n  l1 data cache size = %d Kb\n  l2 cache size = %d Kb\n  Present features :%s\n\n", vendorstr, (int)cpu.family, (int)cpu.l1c, (int)cpu.l1d, (int)cpu.l2, cpu.uisfeat);

  if (p->force_core && strcmp("mmx", p->corename) && strcmp("x86", p->corename)) {
    fprintf(stderr, "Warning : the \"%s\" core doesn't exist, using default core selection rules.\n", p->corename);
    p->force_core = false;
  }

  if ((feat = strstr(cpu.uisfeat, " mmx")) && (feat[4] == ' ' || !feat[4])) {
    if (p->force_core) {
      if (!strcmp("mmx", p->corename)) try_crack_pwl = &try_crack_pwl_mmx;
      else if (!strcmp("x86", p->corename)) try_crack_pwl = &try_crack_pwl_x86;
      else try_crack_pwl = &try_crack_pwl_x86;
    } else {
      if (cpu.vendid == INTEL) try_crack_pwl = &try_crack_pwl_mmx; /* faster on celeron / Pii / Piii */
      else try_crack_pwl = &try_crack_pwl_x86;                     /* faster on athlon / duron */
                                                                   /* assume faster on non intel proc too */
    }
  } else {
    if (p->force_core && !strcmp("mmx", p->corename))
      fprintf(stderr, "Warning : your processor doesn't support mmx instructions, using x86 core.\n");
    try_crack_pwl = &try_crack_pwl_x86;                          /* no mmx available */
  }  
}
