/*
    pwl9x "Windows 9x PassWord List reader"
    Copyright (C) 2001 Knispel Guillaume

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

	.section	.rodata
	.align 16
/* intel cache table */
/* First line : index, 0 = end */
/*   2nd line : 1 = l1 code, 2 = l1 data, 3 = l2*/
/*   3rd line : size in kb  (assume 12 kop is 16 kb code, 16 --> 32 and 32 --> 64) */
cti:	.byte	0x6, 0x8, 0xA, 0xC, 0x41, 0x42, 0x43, 0x44, 0x45, 0x66, 0x67, 0x68, 0x70, 0x71, 0x72, 0x79, 0x7A, 0x7B, 0x7C, 0x82, 0x83, 0x84, 0x85, 0
ctt:	.byte	  1,   1,   2,   2,    3,    3,    3,    3,    3,    2,    2,    2,    1,    1,    1,    3,    3,    3,    3,    3,    3,    3,    3
	.align	4
cts:	.int	  8,  16,   8,  16,  128,  256,  512, 1024, 2048,    8,   16,   32,   16,   32,   64,  128,  256,  512, 1024,  256,  512, 1024, 2048	


/* std feat */
fpu:	.string " fpu"
tsc:	.string	" tsc"
cx8:	.string " cx8"
cmov:	.string " cmov"
clfsh:	.string " clfsh"
mmx:	.string " mmx"
fxsr:	.string " fxsr"
sse:	.string " sse"
sse2:	.string " sse2"

/* amd feat */
tdnow:	.string " 3dnow"
e3dn:	.string	" e3dnow"
emmx:	.string	" emmx"

	.align 4
std_feat:
	.int	0, fpu, 4, tsc, 8, cx8, 15, cmov, 19, clfsh, 23, mmx, 24, fxsr, 25, sse, 26, sse2, 0xff

amd_ext_feat:
	.int	31, tdnow, 30, e3dn, 22, emmx, 0xff

.text

#define vendid		0
#define family		4
#define l1c		8
#define	l1d		12
#define	l2		16
#define uisfeat		20

	.align 16
c7:	movzbl	ctt(%edi), %ebp
	cmp	$1, %ebp
	je	sl1c
	cmp	$2, %ebp
	je	sl1d
	cmp	$3, %ebp
	je	sl2
	ret
sl1c:	movl	cts(,%edi, 4), %ebp
	movl	%ebp, l1c(%esi)
	ret
sl1d:	movl	cts(,%edi, 4), %ebp
	movl	%ebp, l1d(%esi)
	ret
sl2:	movl	cts(,%edi, 4), %ebp
	movl	%ebp, l2(%esi)
	ret

#define INTEL	0x090c3631
#define AMD	0xefe3fb3d
#define IDT	0x181d3637
#define NG	0x3d160b04
#define CYRIX	0x2a0d0134
#define UMC	0xa8fe9fec
#define RISE	0x073802d6
#define TRANSM	0xc0d92cd0
	
	.align 16
	.globl idcpu
idcpu:
	pushl	%ebx
	pushl	%ecx
	pushl	%edx
	pushl	%ebp
	pushl	%edi
	pushl	%esi			/* Save registers */

	movl	28(%esp), %esi		/* get return struct ptr */

/* ASSUME NO CACHE */
	movl	$0, l1c(%esi)
	movl	$0, l1d(%esi)
	movl	$0, l2(%esi)
	
/* CHECK CPUID PRESENCE */
	movl	$4, %ebx
check:	pushf
	popl	%eax			/* get flags */
	xorl	$(1<<21), %eax		/* switch cpuid present flag */
	pushl	%eax
	popf				/* and set flags */
	pushf
	popl	%edx			/* then read them back */
	xorl	%edx, %eax		/* check if cpuid present */
	testl	$(1<<21), %eax
	je	cpuid			/* yes */
	decl	%ebx
	jnz	check			/* do it 4 times cause interrupt handlers could change the flag */
	jmp	no_cpuid		/* cpuid not present */

/* USE CPUID */
cpuid:	
/* VENDOR */
	xorl	%eax, %eax		/* get vendor and max # function */
	cpuid
	roll	$16, %edx
	addl	$0xA5A5A5A5, %ecx
	xorl	%edx, %ebx
	xorl	%ecx, %ebx
	movl	%ebx, vendid(%esi)	/* save mixed vendor id */

	orl	%eax, %eax
	movl	%eax, %ebp		/* save max # function */
	jz	no_family_feat

/* FAMILY / FEATURES */
	movl	$1, %eax		/* family/model/stepping + features */
	cpuid
	shrl	$8, %eax
	andl	$0xF, %eax		/* family */
	movl	%eax, family(%esi)

	leal	uisfeat(%esi), %edi	/* output human readable features string */
	movl	$std_feat, %eax		/* standard features table (bit-->str) */

floop:	movl	(%eax), %ecx		/* bit # */
	cmpl	$0xff, %ecx
	je	fxit			/* 0ffh means no more */
	movl	$1, %ebx
	shll	%cl, %ebx		/* mask */
	testl	%ebx, %edx
	jz	next			/* feature absent */
	movl	4(%eax), %ecx
cp:	movb	(%ecx), %bl		/* copy feature name in output string */
	orb	%bl, %bl
	jz	next			/* dismiss ending 0 */
	movb	%bl, (%edi)
	inc	%ecx			/* update char ptrs */
	inc	%edi
	jmp	cp
next:	addl	$8, %eax		/* update feature ptr */
	jmp	floop
fxit:	movb	$0, (%edi)
	
/* INTEL CACHE */
	cmpl	$ INTEL, vendid(%esi)
	jne	ext_feat_amd_cache

	cmpl	$2, %ebp		/* cache info present ? */
	jnb	intch
	cmpl	$5, family(%esi)
	jne	i486c
	movl	$1, %eax
	cpuid
	testl	$ 1<<23, %edx		/* mmx bit */
	jnz	p5mmx
	movl	$8, l1c(%esi)		/* legacy pentiums have 8 kb l1 code cache, and 8 kb l1 data cache */
	movl	$8, l1d(%esi)
	jmp	endidcpu
p5mmx:	movl	$16, l1c(%esi)		/* mmx pentiums have 16 kb l1 code cache, and 16 kb l1 data cache */
	movl	$16, l1d(%esi)
	jmp	endidcpu
i486c:	cmpl	$4, family(%esi)
	jne	no_cache
	movl	$4, l1c(%esi)		/* 486 have 8 kb unified cache, assume that mean 4kb code & 4kb data */
	movl	$4, l1d(%esi)
	jmp	endidcpu

intch:	movl	$2, %eax
	cpuid
	testl	$0x80000000, %eax
	jnz	tebx
teaxl:	shrl	$8, %eax
	orl	%eax, %eax
	jz	tebx
	movl	$-1, %edi
scanac:	incl	%edi
	cmpb	$0, cti(%edi)
	je	teaxl
	cmpb	cti(%edi), %al
	jne	scanac
	call	c7
	jmp	teaxl

tebx:	testl	$0x80000000, %ebx
	jnz	tecx
tebxl:	orl	%ebx, %ebx
	jz	tecx
	movl	$-1, %edi
scanbc:	incl	%edi
	cmpb	$0, cti(%edi)
	je	iebxl
	cmpb	cti(%edi), %bl
	jne	scanbc
	call	c7
iebxl:	shrl	$8, %ebx
	jmp	tebxl
		
tecx:	testl	$0x80000000, %ecx
	jnz	tedx
tecxl:	orl	%ecx, %ecx
	jz	tedx
	movl	$-1, %edi
scancc:	incl	%edi
	cmpb	$0, cti(%edi)
	je	iecxl
	cmpb	cti(%edi), %cl
	jne	scancc
	call	c7
iecxl:	shrl	$8, %ecx
	jmp	tecxl
	
tedx:	testl	$0x80000000, %edx
	jnz	endidcpu
tedxl:	orl	%edx, %edx
	jz	endidcpu
	movl	$-1, %edi
scandc:	incl	%edi
	cmpb	$0, cti(%edi)
	je	iedxl
	cmpb	cti(%edi), %dl
	jne	scandc
	call	c7
iedxl:	shrl	$8, %edx
	jmp	tedxl

/* ASSUME EXTENDED AMD FEATURES / CACHE AVAILABLE */
ext_feat_amd_cache:
	movl	$0x80000000, %eax
	cpuid
	movl	%eax, %ebp		/* save max ext function */
	movl	$0x80000001, %eax
	cmpl	%eax, %ebp
	jb	amdnch

	cpuid
	movl	$amd_ext_feat, %eax	/* extended features table (bit-->str) */

efloop:	movl	(%eax), %ecx		/* bit # */
	cmpl	$0xff, %ecx
	je	efxit			/* 0ffh means no more */
	movl	$1, %ebx
	shll	%cl, %ebx		/* mask */
	testl	%ebx, %edx
	jz	enext			/* feature absent */
	movl	4(%eax), %ecx
ecp:	movb	(%ecx), %bl		/* copy feature name in output string */
	orb	%bl, %bl
	jz	enext			/* dismiss ending 0 */
	movb	%bl, (%edi)
	inc	%ecx			/* update char ptrs */
	inc	%edi
	jmp	ecp
enext:	addl	$8, %eax		/* update feature ptr */
	jmp	efloop
efxit:	movb	$0, (%edi)

	movl	$0x80000005, %eax
	cmpl	%eax, %ebp
	jnb	amdch

amdnch:	cmpl	$ AMD, vendid(%esi)
	jne	no_cache
	movl	$1, %eax		/* amd k5 model 0 do not have cache info */
	cpuid
	andw	$0xff0, %ax
	cmpw	$0x500, %ax
	jne	no_cache
	movl	$8, l1d(%esi)
	movl	$16, l1c(%esi)
	movl	$0, l2(%esi)
	jmp	endidcpu

amdch:	cpuid
	shrl	$24, %ecx		/* bit 24-->31 of ecx is size of l1 data cache in kb */
	shrl	$24, %edx		/* bit 24-->31 of edx is size of l1 code cache in kb */
	movl	%ecx, l1d(%esi)
	movl	%edx, l1c(%esi)

	cmpl	$ AMD, vendid(%esi)
	jne	nobug
	movl	$1, %eax		/* amd durons step a0 incorrectly report l2 cache size */
	cpuid
	cmpw	$0x630, %ax
	movl	$64, %ecx		/* assume it's 64 kb */
	je	samdl2
nobug:	movl	$0x80000006, %eax
	cmpl	%eax, %ebp
	jb	endidcpu
	cpuid
	shrl	$16, %ecx		/* bit 16-->31 of ecx is size of l2 cache in kb */
samdl2:	movl	%ecx, l2(%esi)

	jmp	endidcpu

no_family_feat:
	movl	$0xFFFFFFFF, family(%esi)
	movb	$0, uisfeat(%esi)

no_cache:
	movl	$0xFFFFFFFF, l1c(%esi)
	movl	$0xFFFFFFFF, l1d(%esi)
	movl	$0xFFFFFFFF, l2(%esi)
	jmp	endidcpu

/* 386 OR 486 ? */
no_cpuid:
	movl	$0xFFFFFFFF, vendid(%esi)
	movl	$3, family(%esi)
	movb	$0, uisfeat(%esi)
	pushf				/* push original EFLAGS */
	popl	%eax			/* get original EFLAGS */
	movl	%eax, %ecx		/* save original EFLAGS */
	xorl	$0x40000, %eax		/* flip AC bit in EFLAGS */
	pushl	%eax			/* save new EFLAGS value on stack */
	popf				/* replace current EFLAGS value */
	pushf				/* get new EFLAGS */
	popl	%eax			/* store new EFLAGS in EAX */
	xorl	%ecx, %eax		/* can't toggle AC bit, processor=80386 */
	movl	$3, family(%esi)
	jz	endidcpu		/* jump if 80386 processor */
	pushl	%ecx
	popf				/* restore AC bit in EFLAGS first */
	movl	$4, family(%esi)	/* it's a 486 */
	movl	$4, l1c(%esi)		/* 486 have 8 kb unified cache, assume that mean 4kb code & 4kb data */
	movl	$4, l1d(%esi)


endidcpu:
	movl	family(%esi), %eax	/* return family */
	popl	%esi
	popl	%edi
	popl	%ebp
	popl	%edx
	popl	%ecx
	popl	%ebx			/* restore registers */
	ret				/* return */
