/*
    pwl9x "Windows 9x PassWord List reader"
    Copyright (C) 2001 Knispel Guillaume

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"
#include "list.h"

static void init_xor_table(xor_table *x, hres k)     /* initializing cryptotable */
{
  int i,p,t,tt; x->c1=0,x->c2=0;
  for (i=0;i<=255;x->body[i]=i,i++);
  t=0; for (i=0,p=0;i<=255;i++,p=i & 0x0F) {
    t+=k[p]+x->body[i]; t&=0xFF;
    tt=x->body[i]; x->body[i]=x->body[t]; x->body[t]=tt;
  }
}

static void use_xor_table(xor_table* x, void *data_in, void *data_out, dword datalen)
{
  int p1,p2,t;          /* en/decrypting the data via cryptotable */
  p1=x->c1; p2=x->c2;
  while (datalen--) {
    p2+=x->body[(++p1, p1 &= 0xff)]; p2 &=0xFF;
    t=x->body[p1]; x->body[p1]=x->body[p2]; x->body[p2]=t;
    t+=x->body[p1]; t&=0xff; *((byte*)data_out)++ = *((byte*)data_in)++ ^ x->body[t];
  };
  x->c1=p1; x->c2=p2;
}

static inline dword SwaBits(dword Parm, byte Where)      /* Swap Bits into Dword */
{
  return (Parm << Where) | (Parm >> (0x20 - Where));
}

static void init_hash(hash_container* h) {      /* hashing initialization */
 h->idx[0]=0; h->idx[1]=0;
 h->cnt[0]=0x67452301; h->cnt[1]=0xEFCDAB89;
 h->cnt[2]=0x98BADCFE; h->cnt[3]=0x10325476;
}

static void calc_hash(dword c[4], dword k[16])   /* main hashing proc. */
{
  dword p1,p2,p3,p4;

  p1=c[0]; p2=c[1]; p3=c[2]; p4=c[3];

/*-----------------------------------------------------------------*/

  p1+=((~p2 & p4) | (p2 & p3))+k[0x0]-0x28955b88; p1=SwaBits(p1,0x07)+p2;
  p4+=((~p1 & p3) | (p1 & p2))+k[0x1]-0x173848AA; p4=SwaBits(p4,0x0C)+p1;
  p3+=((~p4 & p2) | (p4 & p1))+k[0x2]+0x242070DB; p3=SwaBits(p3,0x11)+p4;
  p2+=((~p3 & p1) | (p3 & p4))+k[0x3]-0x3E423112; p2=SwaBits(p2,0x16)+p3;

  p1+=((~p2 & p4) | (p2 & p3))+k[0x4]-0x0A83F051; p1=SwaBits(p1,0x07)+p2;
  p4+=((~p1 & p3) | (p1 & p2))+k[0x5]+0x4787C62A; p4=SwaBits(p4,0x0C)+p1;
  p3+=((~p4 & p2) | (p4 & p1))+k[0x6]-0x57CFB9ED; p3=SwaBits(p3,0x11)+p4;
  p2+=((~p3 & p1) | (p3 & p4))+k[0x7]-0x02B96AFF; p2=SwaBits(p2,0x16)+p3;

  p1+=((~p2 & p4) | (p2 & p3))+k[0x8]+0x698098D8; p1=SwaBits(p1,0x07)+p2;
  p4+=((~p1 & p3) | (p1 & p2))+k[0x9]-0x74BB0851; p4=SwaBits(p4,0x0C)+p1;
  p3+=((~p4 & p2) | (p4 & p1))+k[0xA]-0x0000A44F; p3=SwaBits(p3,0x11)+p4;
  p2+=((~p3 & p1) | (p3 & p4))+k[0xB]-0x76A32842; p2=SwaBits(p2,0x16)+p3;

  p1+=((~p2 & p4) | (p2 & p3))+k[0xC]+0x6B901122; p1=SwaBits(p1,0x07)+p2;
  p4+=((~p1 & p3) | (p1 & p2))+k[0xD]-0x02678E6D; p4=SwaBits(p4,0x0C)+p1;
  p3+=((~p4 & p2) | (p4 & p1))+k[0xE]-0x5986BC72; p3=SwaBits(p3,0x11)+p4;
  p2+=((~p3 & p1) | (p3 & p4))+k[0xF]+0x49B40821; p2=SwaBits(p2,0x16)+p3;

/*-----------------------------------------------------------------*/

  p1+=((~p4 & p3) | (p4 & p2))+k[0x1]-0x09E1DA9E; p1=SwaBits(p1,0x05)+p2;
  p4+=((~p3 & p2) | (p3 & p1))+k[0x6]-0x3FBF4CC0; p4=SwaBits(p4,0x09)+p1;
  p3+=((~p2 & p1) | (p2 & p4))+k[0xB]+0x265E5A51; p3=SwaBits(p3,0x0E)+p4;
  p2+=((~p1 & p4) | (p1 & p3))+k[0x0]-0x16493856; p2=SwaBits(p2,0x14)+p3;

  p1+=((~p4 & p3) | (p4 & p2))+k[0x5]-0x29D0EFA3; p1=SwaBits(p1,0x05)+p2;
  p4+=((~p3 & p2) | (p3 & p1))+k[0xA]+0x02441453; p4=SwaBits(p4,0x09)+p1;
  p3+=((~p2 & p1) | (p2 & p4))+k[0xF]-0x275E197F; p3=SwaBits(p3,0x0E)+p4;
  p2+=((~p1 & p4) | (p1 & p3))+k[0x4]-0x182C0438; p2=SwaBits(p2,0x14)+p3;

  p1+=((~p4 & p3) | (p4 & p2))+k[0x9]+0x21E1CDE6; p1=SwaBits(p1,0x05)+p2;
  p4+=((~p3 & p2) | (p3 & p1))+k[0xE]-0x3CC8F82A; p4=SwaBits(p4,0x09)+p1;
  p3+=((~p2 & p1) | (p2 & p4))+k[0x3]-0x0B2AF279; p3=SwaBits(p3,0x0E)+p4;
  p2+=((~p1 & p4) | (p1 & p3))+k[0x8]+0x455A14ED; p2=SwaBits(p2,0x14)+p3;

  p1+=((~p4 & p3) | (p4 & p2))+k[0xD]-0x561C16FB; p1=SwaBits(p1,0x05)+p2;
  p4+=((~p3 & p2) | (p3 & p1))+k[0x2]-0x03105C08; p4=SwaBits(p4,0x09)+p1;
  p3+=((~p2 & p1) | (p2 & p4))+k[0x7]+0x676F02D9; p3=SwaBits(p3,0x0E)+p4;
  p2+=((~p1 & p4) | (p1 & p3))+k[0xC]-0x72D5B376; p2=SwaBits(p2,0x14)+p3;

/*-----------------------------------------------------------------*/

  p1+=(p4 ^ p3 ^ p2)+k[0x5]-0x0005C6BE; p1=SwaBits(p1,0x04)+p2;
  p4+=(p3 ^ p2 ^ p1)+k[0x8]-0x788E097F; p4=SwaBits(p4,0x0B)+p1;
  p3+=(p4 ^ p2 ^ p1)+k[0xB]+0x6D9D6122; p3=SwaBits(p3,0x10)+p4;
  p2+=(p4 ^ p3 ^ p1)+k[0xE]-0x021AC7F4; p2=SwaBits(p2,0x17)+p3;

  p1+=(p4 ^ p3 ^ p2)+k[0x1]-0x5B4115BC; p1=SwaBits(p1,0x04)+p2;
  p4+=(p3 ^ p2 ^ p1)+k[0x4]+0x4BDECFA9; p4=SwaBits(p4,0x0B)+p1;
  p3+=(p4 ^ p2 ^ p1)+k[0x7]-0x0944B4A0; p3=SwaBits(p3,0x10)+p4;
  p2+=(p4 ^ p3 ^ p1)+k[0xA]-0x41404390; p2=SwaBits(p2,0x17)+p3;

  p1+=(p4 ^ p3 ^ p2)+k[0xD]+0x289B7EC6; p1=SwaBits(p1,0x04)+p2;
  p4+=(p3 ^ p2 ^ p1)+k[0x0]-0x155ED806; p4=SwaBits(p4,0x0B)+p1;
  p3+=(p4 ^ p2 ^ p1)+k[0x3]-0x2B10CF7B; p3=SwaBits(p3,0x10)+p4;
  p2+=(p4 ^ p3 ^ p1)+k[0x6]+0x04881D05; p2=SwaBits(p2,0x17)+p3;

  p1+=(p4 ^ p3 ^ p2)+k[0x9]-0x262B2FC7; p1=SwaBits(p1,0x04)+p2;
  p4+=(p3 ^ p2 ^ p1)+k[0xC]-0x1924661B; p4=SwaBits(p4,0x0B)+p1;
  p3+=(p4 ^ p2 ^ p1)+k[0xF]+0x1FA27CF8; p3=SwaBits(p3,0x10)+p4;
  p2+=(p4 ^ p3 ^ p1)+k[0x2]-0x3B53A99B; p2=SwaBits(p2,0x17)+p3;

/*-----------------------------------------------------------------*/

  p1+=((~p4 | p2) ^ p3)+k[0x0]-0x0BD6DDBC; p1=SwaBits(p1,0x06)+p2;
  p4+=((~p3 | p1) ^ p2)+k[0x7]+0x432AFF97; p4=SwaBits(p4,0x0A)+p1;
  p3+=((~p2 | p4) ^ p1)+k[0xE]-0x546BDC59; p3=SwaBits(p3,0x0F)+p4;
  p2+=((~p1 | p3) ^ p4)+k[0x5]-0x036C5FC7; p2=SwaBits(p2,0x15)+p3;

  p1+=((~p4 | p2) ^ p3)+k[0xC]+0x655B59C3; p1=SwaBits(p1,0x06)+p2;
  p4+=((~p3 | p1) ^ p2)+k[0x3]-0x70F3336E; p4=SwaBits(p4,0x0A)+p1;
  p3+=((~p2 | p4) ^ p1)+k[0xA]-0x00100B83; p3=SwaBits(p3,0x0F)+p4;
  p2+=((~p1 | p3) ^ p4)+k[0x1]-0x7A7BA22F; p2=SwaBits(p2,0x15)+p3;

  p1+=((~p4 | p2) ^ p3)+k[0x8]+0x6FA87E4F; p1=SwaBits(p1,0x06)+p2;
  p4+=((~p3 | p1) ^ p2)+k[0xF]-0x01D31920; p4=SwaBits(p4,0x0A)+p1;
  p3+=((~p2 | p4) ^ p1)+k[0x6]-0x5CFEBCEC; p3=SwaBits(p3,0x0F)+p4;
  p2+=((~p1 | p3) ^ p4)+k[0xD]+0x4E0811A1; p2=SwaBits(p2,0x15)+p3;

  p1+=((~p4 | p2) ^ p3)+k[0x4]-0x08AC817E; p1=SwaBits(p1,0x06)+p2;
  p4+=((~p3 | p1) ^ p2)+k[0xB]-0x42C50DCB; p4=SwaBits(p4,0x0A)+p1;
  p3+=((~p2 | p4) ^ p1)+k[0x2]+0x2AD7D2BB; p3=SwaBits(p3,0x0F)+p4;
  p2+=((~p1 | p3) ^ p4)+k[0x9]-0x14792C6F; p2=SwaBits(p2,0x15)+p3;

/*-----------------------------------------------------------------*/

  c[0]+=p1; c[1]+=p2; c[2]+=p3; c[3]+=p4;
}

static void add_hash(hash_container* cont, void* data, dword datalen)
{
  dword p;           /* add some data to hash */

  p=(cont->idx[0]>>3) & 0x3F;
  /*  if ((datalen<<3)+cont->idx[0] < cont->idx[0]) cont->idx[1]++;
      cont->idx[0]+=datalen<<3; cont->idx[1]+=datalen >> 0x1D;*/ /* we never push 512 mo of data :) */
  cont->idx[0]+=datalen<<3;

  while (datalen--) {
    cont->buf[p]=*((byte*)data)++;
    if (++p==0x40) {
      calc_hash(cont->cnt, (dword *) cont->buf);
      p=0;
    }
  }
}

static byte zero[0x40]={0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,};
static void flush_hash(hash_container* cont)    /* drop buffer & make hash */
{
  dword p1,p2,t[2];

  t[0]=cont->idx[0]; t[1]=cont->idx[1];
  p1=(cont->idx[0]>>3) & 0x3F;
  if (p1<0x38) p2=0x38-p1; else p2=0x78-p1;
  add_hash(cont,zero,p2);

  memcpy(cont->buf+0x38, t, sizeof(dword) * 2);
  calc_hash(cont->cnt,(dword *)cont->buf);
}

static void make_cryption_table(xor_table* t, char *name, char *pwd, dword Seed1, dword Seed2)
{
  hash_container h1, h2;

  init_hash(&h1);                    /* main crypto maker procedure */
  add_hash(&h1,&Seed1,sizeof(Seed1));
  add_hash(&h1,name,strlen(name)+1);
  add_hash(&h1,&Seed2,sizeof(Seed2));
  flush_hash(&h1);

  init_hash(&h2);
  add_hash(&h2,pwd,strlen(pwd)+1);
  add_hash(&h2,h1.cnt,sizeof(h1.cnt));
  flush_hash(&h2);

  init_xor_table(t, (byte *) h2.cnt);
}

static int check_cryption_sign(char* name, hres CryptoSign, hres CheckSign)
{
  hash_container h;                  /* checking decryption legacy */

  init_hash(&h); add_hash(&h,name,strlen(name)+1);
  add_hash(&h,CryptoSign,sizeof(hres)); flush_hash(&h);
  return memcmp(h.cnt,CheckSign,sizeof(hres));
}

void try_list(Params *p, Pwl *pwl)
{
  xor_table x;
  word ResEntry[0x11];
  word ResSz[0x10];
  int i, tr0, tr1, j, jj, k, kk;
  char s[256]; char type[5];

  printf("Content\n---\n\n");

  make_cryption_table(&x,p->username,p->pwd,-1,pwl->head->salt[0x10]);
  use_xor_table(&x,pwl->head->sign,pwl->head->sign,2*sizeof(hres)+sizeof(pwl->head->sectionptr));

  if (check_cryption_sign(p->username,pwl->head->sign,pwl->head->check)) {
    fprintf(stderr, "Sorry, invalid UserName or Password\n");
    exit(EPWD);
  }

  ResEntry[0]=0x290; ResEntry[0x10]=p->size;
  for (i=0;i<0x0F;i++) ResEntry[i+1]=pwl->head->sectionptr[i];
  for (i=0;i<0x10;i++) ResSz[i]=ResEntry[i+1]-ResEntry[i];
  for (i=0;i<0x10;i++) {
    make_cryption_table(&x,p->username,p->pwd,i,pwl->head->salt[i]);
    use_xor_table(&x,pwl->stream+ResEntry[i],pwl->stream+ResEntry[i],ResSz[i]);
  }

  printf("-[Type]-[The resource location string]--------------[Password]-\n");
  tr0=0,tr1=0; for (i=0;i<0x10;i++) if (ResSz[i]>8) {
    j=ResEntry[i]; jj=ResEntry[i+1]-2; tr0++;
    while (j<jj) { k=(unsigned char)pwl->stream[j+7];
    printf("  %s  ",k==3?"Link":k==6?"Dial":k==19?"Web ":
           (sprintf(type, "0x%02X", k), type));
    k=(((word)pwl->stream[j+3]) << 8)+pwl->stream[j+2];
    memcpy(s,pwl->stream+j+8,k); s[k]=0; printf(" %-38s",s);
    kk=(((word)pwl->stream[j+5]) << 8)+pwl->stream[j+4];
    memcpy(s,pwl->stream+j+8+k,kk); s[kk]=0; printf("%14s",s);
    j=j+((word)pwl->stream[j+1] << 8)+pwl->stream[j+0];
    tr1++; printf("\n");
    }
  }
  printf("---------------------------------------------------------------\n");
  printf("  Key Entrys: %d;   Number of resources: %d.\n",tr0,tr1);
}
