/*
    pwl9x "Windows 9x PassWord List reader"
    Copyright (C) 2001 Knispel Guillaume

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "defs.h"
#include "cpu.h"
#include "mmx.h"
#include "x86.h"
#include "list.h"

#define LNSZ 1024

char *skip(char *s) { while(isspace(*s)) s++; return s; }
char *rskip(char *s) { while(!isspace(*s)) s++; return s; }

int ltf[256], dtf[256];
struct uniupper { int low, up; } *uu = NULL;
int nmbuc = 0;
int nmbrm = 0;
#define GRAN 512

static bool upperload(char *uuf)
{
  FILE *fuu; char line[LNSZ], *scan; int l, u;

  if ((fuu = fopen(uuf, "rt")) == NULL) return false;

  while (fgets(line, LNSZ, fuu) != NULL) {
    if ((scan = strchr(line, '#'))) *scan = 0;
    if ((scan = strchr(line, '\n'))) *scan = 0;
    scan = skip(line);
    if (!*scan) continue;
    if (sscanf(scan, "%x %x", &l, &u) != 2) continue;
    if (nmbuc == nmbrm) {
      if ((uu = realloc(uu, (nmbrm += GRAN) * sizeof(struct uniupper))) == NULL) return false;
    }
    uu[nmbuc].low = l; uu[nmbuc].up = u;
    nmbuc++;
  }

  fclose(fuu);
  return true;
}

static bool rdttf(int unimap[256], char *ttfn)
{
  FILE *ftbl; int i, j, k;
  char line[LNSZ], *scan;
  if ((ftbl = fopen(ttfn, "rt")) == NULL) return false;

  for (i = 0; i < 256; i++) unimap[i] = 0xFFFF; /* illegal in unicode */

  while (fgets(line, LNSZ, ftbl) != NULL) {
    if ((scan = strchr(line, '#'))) *scan = 0;
    if ((scan = strchr(line, '\n'))) *scan = 0;
    scan = skip(line);
    if (!*scan) continue;
    if (sscanf(scan, "%i", &j)) {
      scan = skip(rskip(scan));
      if (!sscanf(scan, "%i", &k)) k = 0xFFFF;
      unimap[j] = k;
    }
  }

  fclose(ftbl);
  return true;
}

static bool process_charset_init(Params *p)
{
  int i;

  if (p->load_ltf) {
    if (!rdttf(ltf, p->ltf_file)) {
      fprintf(stderr, "Can't load input translation table file \"%s\"\nUsing iso-8859-1 instead.\n", p->ltf_file);
      for (i = 0; i < 256; i++) ltf[i] = i;
    }
  } else for (i = 0; i < 256; i++) ltf[i] = i; /* iso-8859-1 2 unicode trans table data */

  if (p->load_dtf) {
    if (!rdttf(dtf, p->dtf_file)) {
      fprintf(stderr, "Can't load ouput translation table file \"%s\"\nUsing iso-8859-1 instead.\n", p->dtf_file);
      for (i = 0; i < 256; i++) dtf[i] = i;
    }
  } else for (i = 0; i < 256; i++) dtf[i] = i; /* iso-8859-1 2 unicode trans table data */

  ltf[0] = 0; /* Make sure uniline is in "UnicodeZ" */
  dtf[0] = 0; /* output line in "CpZ" :) */

  if (p->do_upper) upperload(p->upper_file);

  return true;
}

static int uniupper(int unichar)
{
  int i;

  for(i = 0; i < nmbuc; i++) {
    if (uu[i].low == unichar) return uu[i].up;
  }

  return unichar;
}

static void strconv(Params *p, char *s, int in[256], int out[256])
{
  int i, j;
  int *uniline = malloc(sizeof(int)*(strlen(s)+1));

  if (!uniline) return;

  for (i = 0; (uniline[i] = in[(unsigned char) s[i]]); i++);
  if (p->do_upper) for (i = 0; uniline[i]; i++) uniline[i] = uniupper(uniline[i]);
  for (i = 0; uniline[i]; i++) {
    for (j = 0; j < 256; j++) if (uniline[i] == out[j]) { s[i] = j; break; }
  }
}

static int load_raw_pwl(Params *p, Pwl *pwl)
{
  FILE *fp; int a;

  if (!(fp = fopen(p->pwl_file, "rb"))) return 0;
  fseek(fp, 0, SEEK_END); /* this way we make next ftell return the file size */
  if (!(pwl->stream = malloc(p->size = ftell(fp)))) { fprintf(stderr, "Could not allocate enough memory.\n"); exit(EMEM); }
  rewind(fp);

  a = fread(pwl->stream, p->size, 1, fp);
  fclose(fp);
  if (!a || pwl->head->magic != 0x968582E3 || pwl->head->signcheckOffs != 0x252) { free(pwl->stream); return 0; }
  return 1;
}

static void try_extract_username(Params *p, Pwl *pwl)
{
#define LINESIZE 1024
  char line[LINESIZE], *fnp, *fnpu, *v;
  FILE *s = fopen(p->sysini_file, "rb");

  for(fnp = p->pwl_file; *fnp; fnp++); for(; fnp > p->pwl_file && *fnp != '/' && *fnp != '\\'; fnp--); if (*fnp == '/' || *fnp == '\\') fnp++;
  if (!(fnpu = malloc(strlen(fnp)+1))) { fprintf(stderr, "Could not allocate enough memory.\n"); exit(EMEM); }
  strcpy(fnpu, fnp); fnp = fnpu; if (p->do_upper) strconv(p, fnp, dtf, dtf);
  if (!s) {
    fprintf(stderr, "Unable to open the sysini windows file !\n");
    free(fnpu); return;
  }
  while((v = fgets(line, LINESIZE, s))) {
    if (!strncmp("[Password Lists]", line, strlen("[Password Lists]"))) break;
  }
  if (!v) {
    fprintf(stderr, "Unable to find the %s section !\n", "[Password Lists]");
    free(fnpu); return;
  }
  while((v = fgets(line, LINESIZE, s))) {
    if (p->do_upper) strconv(p, line, dtf, dtf);
    for(; *v; v++); for(; v > line && *v != '\\'; v--); if (*v == '\\') v++;
    if (!strncmp(fnp, v, strlen(fnp))) {
      v = strchr(line, '=');
      if (!v) { free(fnpu); return; }
      *v = 0;
      if (!(p->username = malloc(strlen(line)+1))) { fprintf(stderr, "Could not allocate enough memory.\n"); exit(EMEM); }
      strcpy(p->username, line);
      p->have_username = true;
      free(fnpu); return;
    }
    if (v[0] == '[') {
      fprintf(stderr, "No name match !\n");
      free(fnpu); return;
    }
  }

  free(fnpu); return;
}

static void disp_head(Params *p, Pwl *pwl)
{
  int i;
  printf("Header\n---\n\n");

  printf("Variable\tContent\n");

  printf("Magic\t\t0x%X\n", (int) pwl->head->magic);
  printf("Unknown C\t0x%X\n", (int) pwl->head->unknown_d);
  printf("Links\t\t"); for (i=0; i < 0x100; i++) printf("0x%X\n\t\t", (int) pwl->head->unusedlinks[i]); printf("\n");
  printf("Keys\t\t");  for (i=0; i < 0x100; i++) printf("0x%X\n\t\t", (int) pwl->head->unusedkeys[i]); printf("\n");
  printf("ChkPwdOff\t0x%X\n", (int) pwl->head->signcheckOffs);
  printf("Salts\t\t"); for (i=0; i < 0x11; i++) printf("0x%X\n\t\t", (int) pwl->head->salt[i]); printf("\n");
  printf("Unknown A\t0x%X\n\n", (int) pwl->head->unknown_w);
}

static void help()
{
  fprintf(stderr, "Usage\n---\n\n"\
	          "pwl9x action [options] pwlfile.pwl\n"\
	          "action : [-l] [-t] [-c]\n"\
	          "\t-l : list ressources of the given pwl file\n"\
	          "\t-t : display the header\n"\
	          "\t-c : crack (test lots of pwd)\n"\
	          "options : [-p PASSWORD] [-u USERNAME] [-s system.ini]\n"\
                  "          [-m methode] [-d dictionary] [-a alphabet]\n"\
	          "          [-z maxpwdsz] [-o localtransfile] [-n dostransfile]\n"\
	          "          [-b UniUpper] [-f (mmx|x86)]\n\n");
}

int main(int argc, char *argv[])
{
  Params p;
  Pwl pwl;
  int opt = 0;

  memset(&p, 0, sizeof(p));

  fprintf(stderr, "Windows 9x PassWord List reader\n"\
	          "[pwl9x v0.07] (C) 2001 Xilun.\n\n");

  /* COMMAND LINE PARSING */
  p.maxpwdlen = MAXPOSSIBLEPWDLEN;
  while (opt != '?' && (opt = getopt(argc, argv, "ltcp:m:d:a:u:s:z:o:n:b:f:")) != EOF) {
    switch (opt) {
    case 'l' : p.have_action = true; p.do_list = true; break;
    case 't' : p.have_action = true; p.do_head = true; break;
    case 'c' : p.have_action = true; p.do_crack = true; break;
    case 'p' : p.have_pwd = true; p.pwd = optarg; break;
    case 'm' : p.have_methode = true; p.methode = optarg; break;
    case 'u' : p.have_username = true; p.username = optarg; break;
    case 'd' : p.have_dico = true; p.dico_file = optarg; break;
    case 'a' : p.have_alphabet = true; p.abc_file = optarg; break;
    case 's' : p.have_sysini = true; p.sysini_file = optarg; break;
    case 'z' : p.maxpwdlen = atoi(optarg); break;
    case 'o' : p.load_ltf = true; p.ltf_file = optarg; break;
    case 'n' : p.load_dtf = true; p.dtf_file = optarg; break;
    case 'b' : p.do_upper = true; p.upper_file = optarg; break;
    case 'f' : p.force_core = true; p.corename = optarg; break;
    case '?' : help(); exit(ECMDLINE); break;
    default  :
      fprintf(stderr, "Error in getopt output parsing\n");
      exit(ECMDLINE); break;
    }
  }

  if (p.maxpwdlen > MAXPOSSIBLEPWDLEN) {
    p.maxpwdlen = MAXPOSSIBLEPWDLEN;
    fprintf(stderr, "Warning : the maximum password size was above %d and has been decrease.\n", MAXPOSSIBLEPWDLEN);
  }

  if (!p.have_action || !argv[optind] || argv[optind+1]) { help(); exit(ECMDLINE); }
  else p.pwl_file = argv[optind];
  
  /* SERVICE SELECTION */
  process_charset_init(&p);
  if (p.have_username) strconv(&p, p.username, ltf, dtf);
  if (p.have_pwd) strconv(&p, p.pwd, ltf, dtf);
  if (!load_raw_pwl(&p, &pwl)) { fprintf(stderr, "PWL file not loaded or bad PWL file, giving up.\n"); exit(EPWL); }
  if (p.have_sysini) try_extract_username(&p, &pwl); /* the username in system.ini file has precedence over the command line provided username */
  if (p.do_head) disp_head(&p, &pwl);
  if (p.do_list || p.do_crack) {
    if (!p.have_username) { fprintf(stderr, "Username not found, giving up.\n"); exit(ECMDLINE); }
    if (p.do_crack) {
      if (strlen(p.username) > MAXPOSSIBLEPWDLEN) {
	fprintf(stderr, "Sorry, can't bruteforce pwl with a username more than %d char long.\n", MAXPOSSIBLEPWDLEN);
	exit(EPWL);
      }
      chose_crack_core(&p);
      try_crack_pwl(&p, &pwl);
    }
    if (p.do_list) {
      if (!p.have_pwd) { fprintf(stderr, "Password unknown, giving up.\n"); exit(EPWD); }
      try_list(&p, &pwl);
    }
  }

  return 0;
}
