/*
    pwl9x "Windows 9x PassWord List reader"
    Copyright (C) 2001 Knispel Guillaume

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>

#define LNSZ 1024
typedef enum {false, true } bool;

char *skip(char *s) { while(isspace(*s)) s++; return s; }

char *rskip(char *s) { while(!isspace(*s)) s++; return s; }

struct uniupper { int low, up; } *uu = NULL;
int nmbuc = 0;
int nmbrm = 0;
#define GRAN 512

bool upperload(char *uuf)
{
  FILE *fuu; char line[LNSZ], *scan; int l, u;

  if ((fuu = fopen(uuf, "rt")) == NULL) return false;

  while (fgets(line, LNSZ, fuu) != NULL) {
    if ((scan = strchr(line, '#'))) *scan = 0;
    if ((scan = strchr(line, '\n'))) *scan = 0;
    scan = skip(line);
    if (!*scan) continue;
    if (sscanf(scan, "%x %x", &l, &u) != 2) continue;
    if (nmbuc == nmbrm) {
      if ((uu = realloc(uu, (nmbrm += GRAN) * sizeof(struct uniupper))) == NULL) return false;
    }
    uu[nmbuc].low = l; uu[nmbuc].up = u;
    nmbuc++;
  }

  fclose(fuu);
  return true;
}

int uniupper(int unichar)
{
  int i;

  for(i = 0; i < nmbuc; i++) {
    if (uu[i].low == unichar) return uu[i].up;
  }

  return unichar;
}

bool rdttf(int unimap[256], char *ttfn)
{
  FILE *ftbl; int i, j, k;
  char line[LNSZ], *scan;
  if ((ftbl = fopen(ttfn, "rt")) == NULL) return false;

  for (i = 0; i < 256; i++) unimap[i] = 0xFFFF; /* illegal in unicode */

  while (fgets(line, LNSZ, ftbl) != NULL) {
    if ((scan = strchr(line, '#'))) *scan = 0;
    if ((scan = strchr(line, '\n'))) *scan = 0;
    scan = skip(line);
    if (!*scan) continue;
    if (sscanf(scan, "%i", &j)) {
      scan = skip(rskip(scan));
      if (!sscanf(scan, "%i", &k)) k = 0xFFFF;
      unimap[j] = k;
    }
  }

  fclose(ftbl);
  return true;
}

void help()
{
  fprintf(stderr, "Usage\n---\n\n"\
	          "cpunicp [-u unicode_uppercase_map_file] [-i inputfile] [-o outputfile]\n"\
	          "        [-g input_cp2unicode_translation_file]\n"\
	          "        [-p output_cp2unicode_translation_file]\n"\
	          "If a cp2unicode translation table file isn't provided, then the charset used\n"\
	          "will be iso-8859-1.\n\n");
}

int main(int argc, char *argv[])
{
  FILE *fin, *fout;
  char *input_file = NULL, *output_file = NULL, *ittfn = NULL, *ottfn = NULL, *uuf = NULL;
  int i, j, opt = 0, itt[256], ott[256];
  bool have_input = false, have_output = false, upper = false, have_itt = false, have_ott = false;
  char line[LNSZ];
  int uniline[LNSZ];

  fprintf(stderr, "cpunicp is a 8 bit charset translation utility with optional upper casing.\n"\
	          "This utility is part of Windows 9x PassWord List reader\n"\
	          "[pwl9x v0.07] (C) 2001 Xilun.\n\n");
 
  while (opt != '?' && (opt = getopt(argc, argv, "u:i:o:g:p:")) != EOF) {
    switch (opt) { /* parse cmd line */
    case 'u': upper = true; uuf = optarg; break;
    case 'i': have_input = true; input_file = optarg; break; 
    case 'o': have_output = true; output_file = optarg; break;
    case 'g': have_itt = true; ittfn = optarg; break;
    case 'p': have_ott = true; ottfn = optarg; break;
    case '?': help(); exit(1); break;
    default:
      fprintf(stderr, "Error in getopt output parsing\n");
      exit(1); break;
    }
  }

  if (argv[optind]) { help(); return 1;}

  if (!have_input) fin = stdin;
  else if ((fin = fopen(input_file, "rb")) == NULL) {
    fprintf(stderr, "Can't open input file \"%s\" for reading\n\n", input_file);
    return 2;
  }

  if (!have_output) fout = stdout;
  else if ((fout = fopen(output_file, "wb")) == NULL) {
    fprintf(stderr, "Can't open output file \"%s\" for writing\n\n", output_file);
    return 3;
  }

  if (have_itt) {
    if (!rdttf(itt, ittfn)) {
      fprintf(stderr, "Can't load input translation table file \"%s\"\n\n", ittfn);
      return 4;
    }
  } else for (i = 0; i < 256; i++) itt[i] = i; /* iso-8859-1 2 unicode trans table data */

  if (have_ott) {
    if (!rdttf(ott, ottfn)) {
      fprintf(stderr, "Can't load ouput translation table file \"%s\"\n\n", ottfn);
      return 5;
    }
  } else for (i = 0; i < 256; i++) ott[i] = i; /* iso-8859-1 2 unicode trans table data */

  itt[0] = 0; /* Make sure uniline is in "UnicodeZ" */
  ott[0] = 0; /* output line in "CpZ" :) */

  if (upper) upperload(uuf);

  while (fgets(line, LNSZ, fin) != NULL) {
    for (i = 0; (uniline[i] = itt[(unsigned char) line[i]]); i++);
    if (upper) for (i = 0; uniline[i]; i++) uniline[i] = uniupper(uniline[i]);
    for (i = 0; uniline[i]; i++) {
      for (j = 0; j < 256; j++) if (uniline[i] == ott[j]) { line[i] = j; break; }
    }
    fputs(line, fout);
  }

  fclose(fin);
  fclose(fout);
  return 0;
}
