/*
    pwl9x "Windows 9x PassWord List reader"
    Copyright (C) 2001 Knispel Guillaume

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

	.text

#define	p1	%eax
#define p2	%ebx
#define p3	%ecx
#define p4	%edx
#define t1	%esi
#define t2	%edi
	
/* A+=(((C^D)&B)^D)+key[DT]+K; A=SwaBits(A,RK)+B;
   Note : * A+=K+key[DT]; was early executed in previous iteration
	    & so was T=C^D;
	  * (((C^D)&B)^D) == ((~B & D) | (B & C))
	    but ((~B & D) | (B & C)) use one more instruction
	    & have a longer dependency string 
*/
#define MD5_FF(A,B,C,D,T,NT,RK,DT,K) \
	andl	B, T		; \
	movl	C, NT		; \
	xorl	D, T		; \
	addl	$ K, D		; \
	addl	T, A		; \
	addl	DT*4(%ebp), D	; \
	roll	$ RK, A		; \
	xorl	B, NT		; \
	addl	B, A

/* A+=((~D & C) | (D & B))+key[DT]+K; A=SwaBits(A,RK)+B;
   Note : * do not interlace T and NT registers in macro call of MD5_GG
	  * A+=K+key[DT]; was early executed in previous iteration
            & so were T = D; and NT = (~D & C);
	  * ((~D & C) | (D & B)) == (((C^B)&D)^C)
	    the last form uses one less instruction but this doesn't matter
	    on k7 since they have three execution units and the dependency
	    string is shorter on the first form
*/
#define MD5_GG(A,B,C,D,T,NT,RK,DT,K) \
	andl	B, T		; \
	addl	$ K, D		; \
	orl	T, NT		; \
	movl	C, T		; \
	addl	DT*4(%ebp), D	; \
	addl	NT, A		; \
	movl	C, NT		; \
	roll	$ RK, A		; \
	notl	NT		; \
	addl	B, A		; \
	andl	B, NT

/* A+=(D ^ C ^ B)+key[DT]+K; A=SwaBits(A,RK)+B;
   Note : * A+=K+key[DT]; was early executed in previous iteration
            & so was T=D^C;
*/
#define MD5_HH(A,B,C,D,T,NT,RK,DT,K) \
	addl	$ K, D		; \
	xorl	B, T		; \
	addl	DT*4(%ebp), D	; \
	addl	T, A		; \
	movl	C, NT		; \
	roll	$ RK, A		; \
	xorl	B, NT		; \
	addl	B, A

/* A+=((~D | B) ^ C)+key[DT]+K; A=SwaBits(A,RK)+B;
   Note : * A+=K+key[DT]; was early executed in previous iteration
            & so was T=~D;
*/
#define MD5_II(A,B,C,D,T,NT,RK,DT,K) \
	movl	C, NT		; \
	orl	B, T		; \
	addl	$ K, D		; \
	xorl	C, T		; \
	addl	DT*4(%ebp), D	; \
	addl	T, A		; \
	notl	NT		; \
	roll	$ RK, A		; \
	addl	B, A

	.align 16
	.globl x86_h_md5
x86_h_md5:
	pushl	%ecx
	pushl	%edx
        pushl	%ebx
	pushl	%ebp
        pushl	%esi
        pushl	%edi

	movl	28(%esp), %ebp

	movl	$ 0xD76AA477, p1	/* one pass md5 only sub so part of */
	movl	$ 0xF8FA0BCC, p4	/* first & second iterations are */
	addl	(%ebp), p1		/* staticaly computed */
	movl	$ 0xEFCDAB89, p2
	roll	$ 0x07, p1
	addl	4(%ebp), p4
	movl	$ 0x77777777, t2
	addl	p2, p1
	movl	$ 0x98BADCFE, p3

/*	MD5_FF(p1, p2, p3, p4, t1, t2, 0x07,  1, -0x173848AA)*/
	MD5_FF(p4, p1, p2, p3, t2, t1, 0x0C,  2,  0x242070DB)
	MD5_FF(p3, p4, p1, p2, t1, t2, 0x11,  3, -0x3E423112)
	MD5_FF(p2, p3, p4, p1, t2, t1, 0x16,  4, -0x0A83F051)
	MD5_FF(p1, p2, p3, p4, t1, t2, 0x07,  5,  0x4787C62A)
	MD5_FF(p4, p1, p2, p3, t2, t1, 0x0C,  6, -0x57CFB9ED)
	MD5_FF(p3, p4, p1, p2, t1, t2, 0x11,  7, -0x02B96AFF)
	MD5_FF(p2, p3, p4, p1, t2, t1, 0x16,  8,  0x698098D8)
	MD5_FF(p1, p2, p3, p4, t1, t2, 0x07,  9, -0x74BB0851)
	MD5_FF(p4, p1, p2, p3, t2, t1, 0x0C, 10, -0x0000A44F)
	MD5_FF(p3, p4, p1, p2, t1, t2, 0x11, 11, -0x76A32842)
	MD5_FF(p2, p3, p4, p1, t2, t1, 0x16, 12,  0x6B901122)
	MD5_FF(p1, p2, p3, p4, t1, t2, 0x07, 13, -0x02678E6D)
	MD5_FF(p4, p1, p2, p3, t2, t1, 0x0C, 14, -0x5986BC72)
	MD5_FF(p3, p4, p1, p2, t1, t2, 0x11, 15,  0x49B40821)
/*	MD5_FF(p2, p3, p4, p1, t2, t1, 0x16,  1, -0x09E1DA9E)*/
        andl p3, t2	/* FF rewriten with GG early execution instructions */
	movl p4, t1
	xorl p1, t2
	addl $ -0x09E1DA9E, p1
	addl t2, p2
	movl p4, t2
	addl 1*4(%ebp), p1
	notl t2
	roll $ 0x16, p2
	andl p3, t2
	addl p3, p2

	MD5_GG(p1, p2, p3, p4, t1, t2, 0x05,  6, -0x3FBF4CC0)
	MD5_GG(p4, p1, p2, p3, t1, t2, 0x09, 11,  0x265E5A51)
	MD5_GG(p3, p4, p1, p2, t1, t2, 0x0E,  0, -0x16493856)
	MD5_GG(p2, p3, p4, p1, t1, t2, 0x14,  5, -0x29D0EFA3)
	MD5_GG(p1, p2, p3, p4, t1, t2, 0x05, 10,  0x02441453)
	MD5_GG(p4, p1, p2, p3, t1, t2, 0x09, 15, -0x275E197F)
	MD5_GG(p3, p4, p1, p2, t1, t2, 0x0E,  4, -0x182C0438)
	MD5_GG(p2, p3, p4, p1, t1, t2, 0x14,  9,  0x21E1CDE6)
	MD5_GG(p1, p2, p3, p4, t1, t2, 0x05, 14, -0x3CC8F82A)
	MD5_GG(p4, p1, p2, p3, t1, t2, 0x09,  3, -0x0B2AF279)
	MD5_GG(p3, p4, p1, p2, t1, t2, 0x0E,  8,  0x455A14ED)
	MD5_GG(p2, p3, p4, p1, t1, t2, 0x14, 13, -0x561C16FB)
	MD5_GG(p1, p2, p3, p4, t1, t2, 0x05,  2, -0x03105C08)
	MD5_GG(p4, p1, p2, p3, t1, t2, 0x09,  7,  0x676F02D9)
	MD5_GG(p3, p4, p1, p2, t1, t2, 0x0E, 12, -0x72D5B376)
/*	MD5_GG(p2, p3, p4, p1, t1, t2, 0x14,  5, -0x0005C6BE)*/
        andl p3, t1
	addl $ -0x0005C6BE, p1
	orl t1, t2
	addl 5*4(%ebp), p1
	addl t2, p2
	movl p4, t1
	roll $ 0x14, p2
	xorl p3, t1
	addl p3, p2

	MD5_HH(p1, p2, p3, p4, t1, t2, 0x04,  8, -0x788E097F)
	MD5_HH(p4, p1, p2, p3, t2, t1, 0x0B, 11,  0x6D9D6122)
	MD5_HH(p3, p4, p1, p2, t1, t2, 0x10, 14, -0x021AC7F4)
	MD5_HH(p2, p3, p4, p1, t2, t1, 0x17,  1, -0x5B4115BC)
	MD5_HH(p1, p2, p3, p4, t1, t2, 0x04,  4,  0x4BDECFA9)
	MD5_HH(p4, p1, p2, p3, t2, t1, 0x0B,  7, -0x0944B4A0)
	MD5_HH(p3, p4, p1, p2, t1, t2, 0x10, 10, -0x41404390)
	MD5_HH(p2, p3, p4, p1, t2, t1, 0x17, 13,  0x289B7EC6)
	MD5_HH(p1, p2, p3, p4, t1, t2, 0x04,  0, -0x155ED806)
	MD5_HH(p4, p1, p2, p3, t2, t1, 0x0B,  3, -0x2B10CF7B)
	MD5_HH(p3, p4, p1, p2, t1, t2, 0x10,  6,  0x04881D05)
	MD5_HH(p2, p3, p4, p1, t2, t1, 0x17,  9, -0x262B2FC7)
	MD5_HH(p1, p2, p3, p4, t1, t2, 0x04, 12, -0x1924661B)
	MD5_HH(p4, p1, p2, p3, t2, t1, 0x0B, 15,  0x1FA27CF8)
	MD5_HH(p3, p4, p1, p2, t1, t2, 0x10,  2, -0x3B53A99B)
/*	MD5_HH(p2, p3, p4, p1, t2, t1, 0x17,  0, -0x0BD6DDBC)*/
        addl $ -0x0BD6DDBC, p1
	xorl p3, t2
	addl 0*4(%ebp), p1
	addl t2, p2
	movl p4, t1
	roll $ 0x17, p2
	notl t1
	addl p3, p2

	MD5_II(p1, p2, p3, p4, t1, t2, 0x06,  7,  0x432AFF97)
	MD5_II(p4, p1, p2, p3, t2, t1, 0x0A, 14, -0x546BDC59)
	MD5_II(p3, p4, p1, p2, t1, t2, 0x0F,  5, -0x036C5FC7)
	MD5_II(p2, p3, p4, p1, t2, t1, 0x15, 12,  0x655B59C3)
	MD5_II(p1, p2, p3, p4, t1, t2, 0x06,  3, -0x70F3336E)
	MD5_II(p4, p1, p2, p3, t2, t1, 0x0A, 10, -0x00100B83)
	MD5_II(p3, p4, p1, p2, t1, t2, 0x0F,  1, -0x7A7BA22F)
	MD5_II(p2, p3, p4, p1, t2, t1, 0x15,  8,  0x6FA87E4F)
	MD5_II(p1, p2, p3, p4, t1, t2, 0x06, 15, -0x01D31920)
	MD5_II(p4, p1, p2, p3, t2, t1, 0x0A,  6, -0x5CFEBCEC)
	MD5_II(p3, p4, p1, p2, t1, t2, 0x0F, 13,  0x4E0811A1)
	MD5_II(p2, p3, p4, p1, t2, t1, 0x15,  4, -0x08AC817E)
	MD5_II(p1, p2, p3, p4, t1, t2, 0x06, 11, -0x42C50DCB)
	MD5_II(p4, p1, p2, p3, t2, t1, 0x0A,  2,  0x2AD7D2BB)
	MD5_II(p3, p4, p1, p2, t1, t2, 0x0F,  9, -0x14792C6F)
/*	MD5_II(p2, p3, p4, p1, t2, t1, 0x15, 16,  0x00000000)*/
	movl	32(%esp), %ebp
        orl	p3, t2
	addl	$ 0x67452301, p1
	xorl	p4, t2
	addl	$ 0x10325476, p4
	addl	t2, p2
	roll	$ 0x15, p2
	addl	p3, p2
	addl	$ 0x98BADCFE, p3

	addl	$ 0xEFCDAB89, p2
	movl	p1, 0(%ebp)
	movl	p4, 12(%ebp)
	movl	p3, 8(%ebp)
	movl	p2, 4(%ebp)

	popl	%edi
	popl	%esi
	popl	%ebp
	popl	%ebx
	popl	%edx
	popl	%ecx

	ret
